import pytest # noqa
from io import StringIO
import logging
import numba
import numpy as np
from mumott.data_handling import DataContainer
from mumott.methods.basis_sets import SphericalHarmonics
from mumott.methods.projectors import SAXSProjector
from mumott.methods.residual_calculators import GradientResidualCalculator
from mumott.optimization.loss_functions import SquaredLoss
from mumott.optimization.optimizers import LBFGS


def test_minimal_pipeline():
    numba.config.NUMBA_NUM_THREADS = 1
    numba.set_num_threads(1)

    # check that the functionality for deleting projections works
    data_container = DataContainer(data_path='tests/test_half_circle.h5', data_type='h5')
    projections = data_container.projections
    assert len(projections) == 1
    del projections[0]
    assert len(projections) == 0

    # minimal check for append functionality
    dc1 = DataContainer(data_path='tests/test_half_circle.h5', data_type='h5')
    projections1 = dc1.projections
    dc2 = DataContainer(data_path='tests/test_half_circle.h5', data_type='h5')
    projections2 = dc2.projections
    assert len(projections1) == 1
    projection = projections2[0]
    del projections2[0]
    projections1.append(projection)
    assert len(projections1) == 2

    # minimal check for setitem functionality
    dc1 = DataContainer(data_path='tests/test_half_circle.h5', data_type='h5')
    dc2 = DataContainer(data_path='tests/test_half_circle.h5', data_type='h5')
    projection = dc2.projections[0]
    projection.j_offset = 1.243

    assert dc2.geometry.j_offsets[0] == 1.243

    dc2.geometry.k_offsets[0] = 4.214

    assert projection.k_offset == 4.214

    assert dc2.projections.geometry[0] == projection.geometry

    del dc2.projections[0]

    assert projection.j_offset == 1.243
    assert projection.k_offset == 4.214
    projection.data[0][0] = [1, 2, 3]
    dc1.projections[0] = projection

    assert np.all(dc1.projections[0].data[0][0] == [1, 2, 3])
    assert dc1.projections[0].j_offset == 1.243
    assert dc1.projections[0].k_offset == 4.214
    assert dc1.projections[0].geometry == dc1.projections.geometry[0]

    # load data container and check its basic output
    data_container = DataContainer(data_path='tests/test_half_circle.h5', data_type='h5')
    s = str(data_container)
    assert 'DataContainer' in s
    assert 'Corrected for transmission' in s
    assert 'Corrected for transmission' in s
    s = data_container._repr_html_()
    assert 'DataContainer' in s
    assert 'Corrected for transmission' in s
    assert 'Corrected for transmission' in s

    s = str(data_container.geometry)
    assert 'Geometry' in s
    s = data_container.geometry._repr_html_()
    assert 'Geometry' in s
    s = str(data_container.projections)
    assert 'ProjectionStack' in s
    s = data_container.projections._repr_html_()
    assert 'ProjectionStack' in s

    s = str(data_container.projections[0])
    assert 'Projection' in s
    assert 'diode' in s
    assert 'dcf160' in s
    s = data_container.projections[0]._repr_html_()
    assert 'Projection' in s
    assert 'diode' in s
    assert 'dcf160' in s

    bs = SphericalHarmonics(ell_max=6)
    pr = SAXSProjector(data_container.geometry)
    meth = GradientResidualCalculator(data_container, bs, pr)
    lf = SquaredLoss(meth)
    optimizer = LBFGS(lf, maxiter=5)
    for handler in logging.root.handlers[:]:
        logging.root.removeHandler(handler)
    # - send logging to stream
    f = StringIO()
    logging.basicConfig(stream=f, level=logging.INFO)
    # run optimization
    optimizer.optimize()

    # check that coefficients are correct
    coeffs = meth.coefficients
    assert coeffs.size == 1792
    reference_coeffs = np.array([0.80724443, 0.08168501, -0.00411556, 0.31664116, -0.08639619,
                                 0.85544408, 0.14476595, -0.00619784, 0.03969083, -0.00459848,
                                 0.29080056, -0.09653399, 0.41566121, -0.04310696, 0.75111706,
                                 0.19997459, -0.00493211, 0.0758295, -0.01305589, 0.03441943,
                                 -0.00109075, 0.22410809, -0.02289765, 0.36045654, -0.0908057,
                                 0.39344082, -0.02033042, 0.6810509, 0.86794641, 0.07985363,
                                 -0.00881878, 0.48574504, -0.18512889, 0.83626507, 0.13438714,
                                 -0.01848627, 0.0577594, -0.00524878, 0.29769584, -0.11018548])

    print(repr(coeffs.ravel()[:40]))
    assert np.allclose(coeffs.ravel()[:40], reference_coeffs)
