import pytest # noqa
import numpy as np
from mumott.pipelines import (run_mitra, run_sirt, run_group_lasso,
                              run_sigtt, run_phase_matching_alignment, run_discrete_directions)
from mumott.data_handling import DataContainer
from mumott.methods.projectors import SAXSProjector
from mumott.methods.residual_calculators import GradientResidualCalculator
from mumott.methods.basis_sets import SphericalHarmonics, GaussianKernels, TrivialBasis, NearestNeighbor
from mumott.optimization.loss_functions import SquaredLoss, HuberLoss
from mumott.optimization.regularizers import Laplacian, L1Norm
from mumott.optimization.optimizers import GradientDescent, LBFGS
from mumott.methods.utilities.grids_on_the_sphere import football


@pytest.fixture
def data_container():
    return DataContainer('tests/test_half_circle.h5')


def test_mitra_abs(data_container):
    result = run_mitra(data_container, maxiter=2)
    assert result['result']['loss'] == 0
    assert isinstance(result['basis_set'], TrivialBasis)
    assert result['projector']._geometry == data_container.geometry
    assert result['basis_set'].channels == 1


def test_mitra_saxs(data_container):
    result = run_mitra(data_container, use_absorbances=False, maxiter=2)
    assert result['projector']._geometry == data_container.geometry
    assert np.isclose(result['result']['loss'], 1130.3370887993)
    assert isinstance(result['projector'], SAXSProjector)
    assert isinstance(result['residual_calculator'], GradientResidualCalculator)
    assert isinstance(result['basis_set'], GaussianKernels)
    assert isinstance(result['loss_function'], SquaredLoss)
    assert isinstance(result['optimizer'], GradientDescent)


def test_mitra_kwarg(data_container):
    result = run_mitra(data_container,
                       use_absorbances=False,
                       maxiter=2,
                       basis_set_kwargs=dict(grid_scale=12))
    assert isinstance(result['basis_set'], GaussianKernels)
    assert np.isclose(result['result']['loss'], 1849.7302508)
    assert result['basis_set'].grid_scale == 12
    result = run_mitra(data_container,
                       use_absorbances=False,
                       maxiter=2,
                       use_sirt_weights=False)
    assert isinstance(result['basis_set'], GaussianKernels)
    assert result['preconditioner'] is None
    assert np.allclose(result['weights'], data_container.projections.weights)
    assert np.isclose(result['result']['loss'], 5299.55019)


def test_mitra_kwargs(data_container):
    result = run_mitra(data_container,
                       maxiter=3,
                       absorbances=data_container.diode,
                       use_absorbances=False,
                       projector=SAXSProjector,
                       BasisSet=SphericalHarmonics,
                       basis_set_kwargs=dict(ell_max=2),
                       ResidualCalculator=GradientResidualCalculator,
                       residual_calculator_kwargs=dict(use_scalar_projections=False,
                                                       scalar_projections=data_container.diode[..., None]),
                       Regularizers=[dict(name='l1', regularizer=L1Norm(), regularization_weight=1e-3)],
                       LossFunction=HuberLoss,
                       loss_function_kwargs=dict(delta=1e2),
                       Optimizer=GradientDescent,
                       optimizer_kwargs=dict(ftol=1e-1))
    assert isinstance(result['projector'], SAXSProjector)
    assert result['projector']._geometry == data_container.geometry
    assert isinstance(result['residual_calculator'], GradientResidualCalculator)
    assert isinstance(result['basis_set'], SphericalHarmonics)
    assert result['basis_set'].ell_max == 2
    assert not result['residual_calculator']._use_scalar_projections
    assert isinstance(result['loss_function'].regularizers['l1'], L1Norm)
    assert isinstance(result['loss_function'], HuberLoss)
    assert np.isclose(result['loss_function']._delta, 1e2)
    assert isinstance(result['optimizer'], GradientDescent)
    assert np.isclose(result['optimizer']['ftol'], 1e-1)
    assert np.isclose(result['result']['loss'], 10.093005)


def test_sirt(data_container):
    result = run_sirt(data_container, maxiter=2)
    assert result['result']['loss'] == 0
    assert isinstance(result['basis_set'], TrivialBasis)
    assert result['projector']._geometry == data_container.geometry
    assert result['basis_set'].channels == 1
    result = run_sirt(data_container, use_absorbances=False, maxiter=2)
    assert result['projector']._geometry == data_container.geometry
    assert np.isclose(result['result']['loss'], 1073.0314536)
    assert isinstance(result['projector'], SAXSProjector)
    assert isinstance(result['residual_calculator'], GradientResidualCalculator)
    assert isinstance(result['basis_set'], GaussianKernels)
    assert isinstance(result['loss_function'], SquaredLoss)
    assert isinstance(result['optimizer'], GradientDescent)


def test_sirt_kwargs(data_container):
    result = run_sirt(data_container,
                      maxiter=3,
                      absorbances=data_container.diode,
                      use_absorbances=False,
                      projector=SAXSProjector,
                      ResidualCalculator=GradientResidualCalculator)
    assert isinstance(result['projector'], SAXSProjector)
    assert result['projector']._geometry == data_container.geometry
    assert isinstance(result['residual_calculator'], GradientResidualCalculator)
    assert isinstance(result['basis_set'], GaussianKernels)
    assert isinstance(result['optimizer'], GradientDescent)
    assert np.isclose(result['result']['loss'], 633.608189)


def test_sigtt_kwargs(data_container):
    result = run_sigtt(data_container,
                       maxiter=3,
                       Projector=SAXSProjector,
                       BasisSet=SphericalHarmonics,
                       basis_set_kwargs=dict(ell_max=4),
                       ResidualCalculator=GradientResidualCalculator,
                       residual_calculator_kwargs=dict(use_scalar_projections=False,
                                                       scalar_projections=data_container.diode[..., None]),
                       Regularizers=[dict(name='l1', regularizer=L1Norm(), regularization_weight=1e-3)],
                       LossFunction=HuberLoss,
                       loss_function_kwargs=dict(delta=1e2),
                       Optimizer=GradientDescent,
                       optimizer_kwargs=dict(ftol=1e-1))
    assert isinstance(result['projector'], SAXSProjector)
    assert result['projector']._geometry == data_container.geometry
    assert isinstance(result['residual_calculator'], GradientResidualCalculator)
    assert isinstance(result['basis_set'], SphericalHarmonics)
    assert result['basis_set'].ell_max == 4
    assert not result['residual_calculator']._use_scalar_projections
    assert isinstance(result['loss_function'].regularizers['l1'], L1Norm)
    assert isinstance(result['loss_function'], HuberLoss)
    assert np.isclose(result['loss_function']._delta, 1e2)
    assert isinstance(result['optimizer'], GradientDescent)
    assert np.isclose(result['optimizer']['ftol'], 1e-1)
    assert np.isclose(result['result']['loss'], 8.2202103)


def test_sigtt(data_container):
    result = run_sigtt(data_container)
    assert np.isclose(result['result']['fun'], 22.148074)
    assert isinstance(result['projector'], SAXSProjector)
    assert result['projector']._geometry == data_container.geometry
    assert isinstance(result['residual_calculator'], GradientResidualCalculator)
    assert isinstance(result['basis_set'], SphericalHarmonics)
    assert isinstance(result['loss_function'], SquaredLoss)
    assert isinstance(result['loss_function'].regularizers['laplacian'], Laplacian)
    assert isinstance(result['optimizer'], LBFGS)
    assert result['basis_set'].ell_max == 2


def test_discrete_directions(data_container):
    directions = football()
    result = run_discrete_directions(data_container, directions, maxiter=10)
    print(np.sum(result['result']['x']))
    assert np.isclose(np.sum(result['result']['x']), 762.466004)
    assert isinstance(result['basis_set'], NearestNeighbor)


def test_group_lasso_pipeline(data_container):
    result = run_group_lasso(data_container, 5.1e1, ell_max=2, step_size_parameter=0.2, maxiter=10)
    print(np.sum(result['result']['x']))
    assert np.isclose(np.sum(result['result']['x']), -5363201.455342014)
    assert isinstance(result['basis_set'], SphericalHarmonics)
    assert len(result['basis_set']) == 6
    assert isinstance(result['loss_function'], SquaredLoss)
    bs = GaussianKernels(grid_scale=2)
    result = run_group_lasso(data_container, 1e1, maxiter=10, basis_set=bs, enforce_non_negativity=True,
                             step_size_parameter=0.75)
    print(np.sum(result['result']['x']))
    assert np.isclose(np.sum(result['result']['x']), 905.6779133209385)
    assert isinstance(result['basis_set'], GaussianKernels)
    assert np.all(result['result']['x'] >= 0)
    x0 = np.ones((*data_container.geometry.volume_shape, len(bs)))
    result = run_group_lasso(data_container, 1e1, maxiter=10, basis_set=bs, enforce_non_negativity=True,
                             step_size_parameter=0.75, x0=x0)
    print(np.sum(result['result']['x']))
    assert np.isclose(np.sum(result['result']['x']), 905.9760696091718)


def test_phase_matching_alignment_upsampled(data_container, caplog):
    data_container.geometry.j_offsets[0] = 0.001
    data_container.geometry.k_offsets[0] = -0.754
    data_container.projections[0].diode = np.arange(16.).reshape(4, 4)
    run_phase_matching_alignment(
        data_container, use_gpu=False,
        reconstruction_pipeline_kwargs=dict(maxiter=1,
                                            use_absorbances=True),
        maxiter=2, shift_tolerance=0.001, upsampling=20, relative_sample_size=1.,
        relaxation_weight=0., center_of_mass_shift_weight=0.)
    print(data_container.geometry.j_offsets, data_container.geometry.k_offsets)
    assert np.allclose(data_container.geometry.j_offsets, 0.3133475237)
    assert np.allclose(data_container.geometry.k_offsets, -0.363565595278)
    with pytest.raises(ValueError, match='align_j and align_k'):
        run_phase_matching_alignment(data_container, align_j=False, align_k=False)


def test_phase_matching_alignment(data_container, caplog):
    data_container.geometry.j_offsets[0] = 0.001
    data_container.geometry.k_offsets[0] = -0.754
    data_container.projections[0].diode = np.arange(16.).reshape(4, 4)
    run_phase_matching_alignment(data_container, use_gpu=False,
                                 reconstruction_pipeline_kwargs=dict(maxiter=1,
                                                                     use_absorbances=True),
                                 maxiter=2, shift_tolerance=0.001, upsampling=1, relative_sample_size=1.,
                                 relaxation_weight=0., center_of_mass_shift_weight=0.)
    print(data_container.geometry.j_offsets, data_container.geometry.k_offsets)
    assert np.allclose(data_container.geometry.j_offsets, 3.001)
    assert np.allclose(data_container.geometry.k_offsets, 0.246)


def test_phase_matching_alignment_no_k(data_container, caplog):
    data_container.geometry.j_offsets[0] = 0.001
    data_container.geometry.k_offsets[0] = -0.754
    data_container.projections[0].diode = np.arange(16.).reshape(4, 4)
    run_phase_matching_alignment(
        data_container, use_gpu=False,
        reconstruction_pipeline_kwargs=dict(maxiter=1,
                                            use_absorbances=True),
        maxiter=2, shift_tolerance=0.001, upsampling=20, relative_sample_size=1.,
        relaxation_weight=0., center_of_mass_shift_weight=0., align_k=False)
    print(data_container.geometry.j_offsets, data_container.geometry.k_offsets)
    assert np.allclose(data_container.geometry.j_offsets, 0.31334752377)
    assert np.allclose(data_container.geometry.k_offsets, -0.754)


def test_phase_matching_alignment_no_j(data_container, caplog):
    data_container.geometry.j_offsets[0] = 0.534
    data_container.geometry.k_offsets[0] = -0.754
    data_container.projections[0].diode = np.arange(16.).reshape(4, 4)
    run_phase_matching_alignment(data_container, use_gpu=False,
                                 reconstruction_pipeline_kwargs=dict(maxiter=1,
                                                                     use_absorbances=True),
                                 maxiter=2, shift_tolerance=0.001, relative_sample_size=1.,
                                 relaxation_weight=0., center_of_mass_shift_weight=0., align_j=False)
    print(data_container.geometry.j_offsets, data_container.geometry.k_offsets)
    assert np.allclose(data_container.geometry.j_offsets, 0.534)
    assert np.allclose(data_container.geometry.k_offsets, 1.246)
