import pytest
import numpy as np
from numpy import pi
from mumott.pipelines.fbp_utilities import (_get_unique_indices, _get_filter,
                                            _get_orthogonal_axis, get_filtered_projections)
from mumott import Geometry, DataContainer
from mumott.core.geometry import GeometryTuple


@pytest.fixture
def geometry():
    geo = Geometry()
    geo.append(GeometryTuple(rotation=np.eye(3), j_offset=0.3, k_offset=0.7,
                             inner_angle=0.7, outer_angle=0., inner_axis=np.array((0, 0, 1.)),
                             outer_axis=np.array((1, 0, 0.))),)
    return geo


angles = [np.arange(0, 2 * pi, pi / 4)]
indices = [np.arange(angles[0].size)]
expected_output = [[0, 1, 2, 3]]


@pytest.mark.parametrize('angle, index, outp',
                         [t for t in zip(angles, indices, expected_output)]
                         )
def test_get_unique_indices(angle, index, outp):
    assert np.allclose(_get_unique_indices(angle, index), outp)


def test_get_filter():
    with pytest.raises(ValueError, match='Unknown filter'):
        _get_filter(10, 'abs')
    rlf = _get_filter(10, 'Ram-Lak')
    hmf = _get_filter(10, 'Hamming')
    hnf = _get_filter(10, 'Hann')
    hcf = _get_filter(10, 'cosine')
    hsf = _get_filter(10, 'Shepp-Logan')
    print(repr(rlf))
    print(repr(hmf))
    print(repr(hnf))
    print(repr(hsf))
    print(repr(hcf))
    assert np.allclose(rlf, np.array([0., 0.00795775, 0.01591549, 0., 0.,
                                      0., 0., 0., 0.01591549, 0.00795775]))
    assert np.allclose(hmf, np.array([0., 0.00795317, 0.01587892, 0., 0.,
                                      0., 0., 0., 0.01587892, 0.00795317]))
    assert np.allclose(hnf, np.array([0., 0.00795277, 0.01587574, 0., 0.,
                                      0., 0., 0., 0.01587574, 0.00795277]))
    assert np.allclose(hsf, np.array([0., 0.00795692, 0.01590886, 0., 0.,
                                      0., 0., 0., 0.01590886, 0.00795692]))
    assert np.allclose(hcf, np.array([0., 0.00795526, 0.0158956, 0., 0.,
                                      0., 0., 0., 0.0158956, 0.00795526]))


projection_index = 0
axis_string = 'inner'
expected_output = 1


def test_get_orthogonal_axis(geometry):
    print(geometry)
    result = _get_orthogonal_axis(geometry, projection_index, axis_string)
    assert result == 1
    geometry.inner_axes = np.array((1., 0., 0.))
    result = _get_orthogonal_axis(geometry, projection_index, axis_string)
    assert result == 2
    with pytest.raises(ValueError, match='axis_string'):
        _get_orthogonal_axis(geometry, projection_index, 'test')
    geometry.inner_axes = np.array((0.7, 0, 0.7))
    with pytest.raises(ValueError, match='orthogonal'):
        _get_orthogonal_axis(geometry, projection_index, 'inner')


# use print(repr(array)) to get parseable result
filtered_projection = np.array([[0., 0., 0., -0.00129632, -0.00129632,
                                 -0.00129632, -0.00129632, -0.00129632, -0.00129632, -0.00129632,
                                 -0.00129632, -0.00129632, 0., 0., 0.],
                                [0., 0., 0., -0.00152674, -0.00152674,
                                 -0.00152674, -0.00152674, -0.00152674, -0.00152674, -0.00152674,
                                 -0.00152674, -0.00152674, 0., 0., 0.],
                                [0., 0., 0., -0.00106247, -0.00106247,
                                 -0.00106247, -0.00106247, -0.00106247, -0.00106247, -0.00106247,
                                 -0.00106247, -0.00106247, 0., 0., 0.],
                                [0., 0., 0., 0.00039321, 0.00039321,
                                 0.00039321, 0.00039321, 0.00039321, 0.00039321, 0.00039321,
                                 0.00039321, 0.00039321, 0., 0., 0.],
                                [0., 0., 0., 0.00170869, 0.00170869,
                                 0.00170869, 0.00170869, 0.00170869, 0.00170869, 0.00170869,
                                 0.00170869, 0.00170869, 0., 0., 0.],
                                [0., 0., 0., 0.00160115, 0.00160115,
                                 0.00160115, 0.00160115, 0.00160115, 0.00160115, 0.00160115,
                                 0.00160115, 0.00160115, 0., 0., 0.],
                                [0., 0., 0., 0.00036444, 0.00036444,
                                 0.00036444, 0.00036444, 0.00036444, 0.00036444, 0.00036444,
                                 0.00036444, 0.00036444, 0., 0., 0.],
                                [0., 0., 0., -0.0003639, -0.0003639,
                                 -0.0003639, -0.0003639, -0.0003639, -0.0003639, -0.0003639,
                                 -0.0003639, -0.0003639, 0., 0., 0.],
                                [0., 0., 0., 0.00036444, 0.00036444,
                                 0.00036444, 0.00036444, 0.00036444, 0.00036444, 0.00036444,
                                 0.00036444, 0.00036444, 0., 0., 0.],
                                [0., 0., 0., 0.00160115, 0.00160115,
                                 0.00160115, 0.00160115, 0.00160115, 0.00160115, 0.00160115,
                                 0.00160115, 0.00160115, 0., 0., 0.],
                                [0., 0., 0., 0.00170869, 0.00170869,
                                 0.00170869, 0.00170869, 0.00170869, 0.00170869, 0.00170869,
                                 0.00170869, 0.00170869, 0., 0., 0.],
                                [0., 0., 0., 0.00039321, 0.00039321,
                                 0.00039321, 0.00039321, 0.00039321, 0.00039321, 0.00039321,
                                 0.00039321, 0.00039321, 0., 0., 0.],
                                [0., 0., 0., -0.00106247, -0.00106247,
                                 -0.00106247, -0.00106247, -0.00106247, -0.00106247, -0.00106247,
                                 -0.00106247, -0.00106247, 0., 0., 0.],
                                [0., 0., 0., -0.00152674, -0.00152674,
                                 -0.00152674, -0.00152674, -0.00152674, -0.00152674, -0.00152674,
                                 -0.00152674, -0.00152674, 0., 0., 0.],
                                [0., 0., 0., -0.00129632, -0.00129632,
                                 -0.00129632, -0.00129632, -0.00129632, -0.00129632, -0.00129632,
                                 -0.00129632, -0.00129632, 0., 0., 0.]])

filtered_indices = np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16,
                             17, 18, 19, 20])


def test_get_filtered_projections():
    dc = DataContainer('tests/test_fbp_data.h5')
    p, i = get_filtered_projections(dc.projections)
    print(repr(p[0, ..., 0]))
    assert np.allclose(p[0, ..., 0], filtered_projection)
    assert np.allclose(filtered_indices, i)
