import pytest  # noqa
import logging

import numpy as np
from mumott import DataContainer

from mumott.methods.basis_sets import SphericalHarmonics
from mumott.methods.projectors import SAXSProjector
from mumott.methods.residual_calculators import GradientResidualCalculator
from mumott.optimization.regularizers import L1Norm
from mumott.optimization.loss_functions import SquaredLoss
from mumott.optimization.optimizers import LBFGS

LOGGER = logging.getLogger(__name__)

example_coeffs = np.array([[[0.22733602, 0.31675834, 0.79736546, 0.67625467, 0.39110955,
                             0.33281393, 0.59830875, 0.18673419],
                            [0.67275604, 0.94180287, 0.24824571, 0.94888115, 0.66723745,
                             0.09589794, 0.44183967, 0.88647992],
                            [0.6974535, 0.32647286, 0.73392816, 0.22013496, 0.08159457,
                             0.1598956, 0.34010018, 0.46519315]],
                           [[0.26642103, 0.8157764, 0.19329439, 0.12946908, 0.09166475,
                             0.59856801, 0.8547419, 0.60162124],
                            [0.93198836, 0.72478136, 0.86055132, 0.9293378, 0.54618601,
                             0.93767296, 0.49498794, 0.27377318],
                            [0.45177871, 0.66503892, 0.33089093, 0.90345401, 0.25707418,
                             0.33982834, 0.2588534, 0.35544648]],
                           [[0.00502233, 0.62860454, 0.28238271, 0.06808769, 0.61682898,
                             0.17632632, 0.30438839, 0.44088681],
                            [0.15020234, 0.21792886, 0.47433312, 0.47636886, 0.25523235,
                             0.29756527, 0.27906712, 0.26057921],
                            [0.48276159, 0.21197904, 0.4956306, 0.24626133, 0.83848265,
                             0.18013059, 0.86215629, 0.17829944]]])


def test_optimise(caplog):
    caplog.set_level(logging.WARNING)
    dc = DataContainer('tests/test_half_circle.h5')
    bs = SphericalHarmonics()
    pr = SAXSProjector(dc.geometry)
    meth = GradientResidualCalculator(dc, bs, pr)
    lf = SquaredLoss(meth)
    reg = L1Norm()
    lf.add_regularizer('laplacian', reg, 1e-2)
    test_dict = dict(test1=1, disp=False, test2=2, test3=123, maxiter=1)
    opt = LBFGS(lf, **test_dict)
    result = opt.optimize()
    assert 'Unknown option test1' in caplog.text
    assert np.isclose(result['fun'], 5366.9770828475)
    assert np.isclose(result['nit'], 1)


def test_x0(caplog):
    caplog.set_level(logging.WARNING)
    dc = DataContainer('tests/test_half_circle.h5')
    bs = SphericalHarmonics()
    pr = SAXSProjector(dc.geometry)
    meth = GradientResidualCalculator(dc, bs, pr)
    lf = SquaredLoss(meth)
    reg = L1Norm()
    lf.add_regularizer('laplacian', reg, 1e-1)
    test_dict = dict(disp=False, test1=2, test3=123, maxiter=1, x0=np.ones_like(lf.initial_values))
    opt = LBFGS(lf, **test_dict)
    result = opt.optimize()
    assert 'Unknown option test1' in caplog.text
    assert np.isclose(result['fun'], 2933.14230778)
    assert np.isclose(result['nit'], 1)


def test_hash(caplog):
    reg = L1Norm()
    assert hex(hash(reg))[2:8] == '5808b8'
    dc = DataContainer('tests/test_half_circle.h5')
    bs = SphericalHarmonics()
    pr = SAXSProjector(dc.geometry)
    meth = GradientResidualCalculator(dc, bs, pr)
    lf = SquaredLoss(meth)
    assert hex(hash(lf))[2:8] == '1a323b'
    reg = L1Norm()
    lf.add_regularizer('laplacian', reg, 1e-2)
    assert hex(hash(lf))[2:8] == '1b0e11'
    lf.regularization_weights['laplacian'] = 0.0005
    assert hex(hash(lf))[2:8] == '1398d2'


def test_str():
    reg = L1Norm()
    s = str(reg)
    assert '5808b8' in s


def test_html():
    reg = L1Norm()
    s = reg._repr_html_()
    assert '5808b8' in s


def test_flags():
    reg = L1Norm()

    res_1 = reg.get_regularization_norm(example_coeffs, get_gradient=True)
    assert np.isclose(res_1['regularization_norm'], np.sum(np.abs(example_coeffs)))
    assert np.allclose(res_1['gradient'], np.sign(example_coeffs))

    exp_grad = np.sign(example_coeffs)
    exp_grad[..., -2:] = 0
    res_1 = reg.get_regularization_norm(example_coeffs, get_gradient=True, gradient_part='full')
    assert np.isclose(res_1['regularization_norm'], np.sum(np.abs(example_coeffs[..., :-2])))
    assert np.allclose(res_1['gradient'], exp_grad)

    res_1 = reg.get_regularization_norm(example_coeffs, get_gradient=True, gradient_part='coefficients')
    assert np.isclose(res_1['regularization_norm'], np.sum(np.abs(example_coeffs[..., :-2])))
    assert np.allclose(res_1['gradient'], exp_grad)

    res_1 = reg.get_regularization_norm(example_coeffs, get_gradient=True, gradient_part='angles')
    assert np.isclose(res_1['regularization_norm'], np.sum(np.abs(example_coeffs[..., :-2])))
    assert np.allclose(res_1['gradient'], 0.0)
