import numpy as np
from mumott.pipelines.utilities import image_processing as imp


def reference_image(**kwargs):
    """
    create a image of gradient values, allowing nul gradient on X or X with the kwargs

    Parameters
    ----------
    kwargs
        possible kwargs : homogeneous_on_axis

    Returns
    -------
    ref_img : float[21,21]
        the gradient image.


    possible kwargs :
    homogeneous_on_axis : int, force homogeneity on the giver axis

    """

    points = 11
    steps = 2 / (points - 1)
    grad_one_axis = np.arange(-1, 1 + steps, steps)
    X, Y = np.meshgrid(grad_one_axis, grad_one_axis)

    homogeneous_on_axis = kwargs.get('homogeneous_on_axis', -1)
    if homogeneous_on_axis == 0:
        ref_img = X
    elif homogeneous_on_axis == 1:
        ref_img = Y
    else:
        ref_img = X * Y
    return ref_img


def test_compute_tukey_window():
    """
    test the tukey window based on its shape. Might worth to think of a way to test numerical results

    """

    # create the ref image
    ref_img = reference_image()
    shape = ref_img.shape

    # create window
    window = imp.compute_tukey_window(shape[0], shape[1])

    # check dimensions
    assert np.allclose(window.shape, (shape[0], shape[1], 1))


def test_get_img_grad():
    """
    test the gradient calculus for 2D images, shape and consitencies of values

    """

    # create the ref image, with equal gradient on each direction
    ref_img = reference_image()
    # create the ref image, with 0 gradient on X
    ref_img_no_X = reference_image(homogeneous_on_axis=1)
    # create the ref image, with 0 gradient on Y
    ref_img_no_Y = reference_image(homogeneous_on_axis=0)

    # compute gradient per axis
    dX, dY = imp.get_img_grad(ref_img, 1, 0)  # x is on 0 axis
    dX_no_X, _ = imp.get_img_grad(ref_img_no_X, 1, 0)
    _, dY_no_Y = imp.get_img_grad(ref_img_no_Y, 1, 0)

    # check size
    assert np.allclose(dX.shape, ref_img.shape)
    assert np.allclose(dY.shape, ref_img.shape)
    assert np.allclose(dX_no_X.shape, ref_img.shape)
    assert np.allclose(dY_no_Y.shape, ref_img.shape)

    # check values
    # since ref_img = transpose(ref_img) dX must be = transpose(dY)
    assert np.allclose(dX, np.transpose(dY))
    assert np.allclose(dX_no_X, np.zeros(np.shape(ref_img_no_X)))
    assert np.allclose(dY_no_Y, np.zeros(np.shape(ref_img_no_Y)))


def test_imshift_fft():
    """
    test the shift by fourier space of image stack by comparison and reversibility

    """
    # create the ref image, with equal gradient on each direction
    ref_img = reference_image()
    # stack it
    ref_images_stack = np.concatenate((ref_img[..., np.newaxis], ref_img[..., np.newaxis]), axis=2)

    shift_x = np.asarray([1, 2])  # in px
    shift_y = np.asarray([2, 3])
    shifted_image_stack = imp.imshift_fft(ref_images_stack, shift_y, shift_x)

    # check size
    assert np.allclose(ref_images_stack.shape, shifted_image_stack.shape)

    # check value
    middle_point = (round(ref_images_stack.shape[0] / 2), round(ref_images_stack.shape[1] / 2))

    # check that shift has been made in all stack in the 2 dimensions
    for index_stack in range(ref_images_stack.shape[-1]):
        assert np.allclose(
            ref_images_stack[
                middle_point[0] - shift_y[index_stack], middle_point[1] - shift_x[index_stack], index_stack
            ],
            shifted_image_stack[middle_point[0], middle_point[1], index_stack],
        )
    # check reversibility
    back_image_stack = imp.imshift_fft(shifted_image_stack, -shift_y, -shift_x)
    assert np.allclose(ref_images_stack, back_image_stack)


def test_smooth_edges():
    """
    test the egdes smoothing, that smooth high frequency with a blurring kernel, of a stack of images,
    tested on size, average constance and untouch values at the center
    """

    # create the ref image, with equal gradient on each direction
    ref_img = reference_image()
    # stack it
    ref_images_stack = np.concatenate((ref_img[..., np.newaxis], ref_img[..., np.newaxis]), axis=2)

    # apply smooth
    smoothed_image_stack = imp.smooth_edges(ref_images_stack)

    # check size
    assert np.allclose(ref_images_stack.shape, smoothed_image_stack.shape)

    # center should be untouch, only border over the size of the kernel + 1 -- 3 + 1 by default
    smoothing_size = 4
    assert np.allclose(
        ref_images_stack[smoothing_size: -smoothing_size - 1, smoothing_size: -smoothing_size - 1, :],
        smoothed_image_stack[smoothing_size: -smoothing_size - 1, smoothing_size: -smoothing_size - 1, :],
    )

    # average should be unchange
    assert np.allclose(np.mean(np.mean(ref_images_stack, 0), 0), np.mean(np.mean(smoothed_image_stack, 0), 0))


def test_imfilter_high_pass_1d():
    """
    test the imfilter_high_pass_1d, function that remove some low frequencies of one dimension
    tested on size, and tranpose consistencies, not on values
    """

    # create the ref image, with equal gradient on each direction
    ref_img = reference_image()
    # stack it
    ref_images_stack = np.concatenate((ref_img[..., np.newaxis], ref_img[..., np.newaxis]), axis=2)

    # apply high pass filter with default parameters 0.1
    sigma = 0.1
    filtered_images_stack_x = imp.imfilter_high_pass_1d(ref_images_stack, 1, sigma)
    filtered_images_stack_y = imp.imfilter_high_pass_1d(ref_images_stack, 0, sigma)

    # check size
    assert np.allclose(ref_images_stack.shape, filtered_images_stack_x.shape)
    assert np.allclose(ref_images_stack.shape, filtered_images_stack_y.shape)

    # since ref_img = transpose(ref_img) filtered_images_stack_x must be = transpose(filtered_images_stack_y)
    # on the images axis
    assert np.allclose(filtered_images_stack_y, np.transpose(filtered_images_stack_x, (1, 0, 2)))


def test_center():
    """
    test the center function that compute the coordinates of the center of a tomogram and its associated
    weights
    tested on size, and tranpose consistencies, not on values
    """

    ref_img = reference_image()

    # stack it
    stack_size = 5

    ref_images_stack = ref_img[..., np.newaxis]

    for index_stack in range(stack_size - 1):
        ref_images_stack = np.concatenate((ref_images_stack, ref_img[..., np.newaxis]), axis=2)
    # compute center
    pos_x, pos_y, mass = imp.center(
        np.abs(ref_images_stack)
    )  # abs value to have a non null mass, but still centered at the center of each image

    # uniform ref
    pos_x_ref, pos_y_ref, mass_ref = imp.center(np.ones(ref_images_stack.shape))

    # check size
    assert np.allclose(pos_x.shape, stack_size)
    assert np.allclose(pos_y.shape, stack_size)

    # definition check from the function itself :: center(ones(N), use_shift = true) == [0,0]
    assert np.allclose(pos_x_ref, np.zeros(pos_x_ref.shape))
    assert np.allclose(pos_y_ref, np.zeros(pos_y_ref.shape))

    # center should be at the middle due to symmetries
    # so correction vector should be null
    assert np.allclose(pos_x, pos_y)
    assert np.allclose(pos_x, np.zeros(pos_x.shape))
    assert np.allclose(pos_y, np.zeros(pos_y.shape))
