# -*- coding: utf-8 -*-
"""
Created on Mon Oct 23 11:33:12 2023
@author: baroni_a

Test functions for the alignment odstrcil functions, based on the mumott tomographic reconstructions.
"""

import numpy as np

# mumott dependancies
from mumott.data_handling import DataContainer
from mumott.data_handling.utilities import get_absorbances
from mumott.core.geometry import Geometry, GeometryTuple
from mumott.core.projection_stack import Projection
from mumott.pipelines import run_mitra
from mumott.methods.projectors import SAXSProjector

# more dependancies
from mumott.pipelines.utilities import image_processing as imp
from mumott.pipelines.utilities.alignment_geometry import get_alignment_geometry

# lib to test
from mumott.pipelines import optical_flow_alignment as amo


def ref_tomogram(**kwargs):
    """
    create an absorption tomogram of a cylinder. Values controllable by kwargs. Rotation axis is here consider
    as being x aka 1

    Parameters
    ----------
    kwargs
        possible kwargs : geometry

    Returns
    -------
    tomogram : float[y, x, y]
        the absortpion tomogram.


    possible kwargs :
        geometry : the Geometry object related to the tomogram

    """

    # check if geometry is imposed by kwargs, if not create nescessary values
    geometry = kwargs.get('geometry', None)
    if geometry is None:
        length = 21
        thickness = 11
        volume = np.asarray([length, thickness, thickness])  # rotation axis is here consider as being x aka 1
    else:
        volume = geometry.volume_shape
        length = volume[0]
        thickness = volume[-1]
    # the internal shape is a uniform cylinder
    # dimension of the cylinder
    radius = np.floor(thickness / 2) - 1
    height = length - 4

    tomogram = np.zeros(volume)
    # defining disk
    y, x = np.ogrid[
        -np.floor(thickness / 2): np.floor(thickness / 2) + 1,
        -np.floor(thickness / 2): np.floor(thickness / 2) + 1,
    ]
    mask = x**2 + y**2 <= radius**2

    for index_z in range(length):
        if index_z < height:
            tomogram[index_z, :, :] = mask
    # add the extra dim for mumott
    tomogram = tomogram[..., np.newaxis]

    return tomogram


def rotation_matrix_3D(angle, axis):
    """
    create an 3D rotation matrix according to an angle around a given axis. If axis is badly given it will
    produce an error

    Parameters
    ----------
    angle : float
        the angle to rotate in radians
    axis : str
        the axis on wich to revovle

    Returns
    -------
    R : float[3, 3]
        the rotation matrix
    """

    if axis == 'x':
        R = np.array([[1, 0, 0], [0, np.cos(angle), -np.sin(angle)], [0, np.sin(angle), np.cos(angle)]])
    if axis == 'y':
        R = np.array([[np.cos(angle), 0, np.sin(angle)], [0, 1, 0], [-np.sin(angle), 0, np.cos(angle)]])
    if axis == 'z':
        R = np.array([[np.cos(angle), -np.sin(angle), 0], [np.sin(angle), np.cos(angle), 0], [0, 0, 1]])
    return R


def rot_matrices(rotation, tilt):
    """
    create an array of 3D rotation matrix according to a rotation angle and tilt angle, assuming rotation
    around x and tilt around y

    Parameters
    ----------
    rotation : float[*]
        the angles to rotate in radians
    tilt : float[*]
        the angles to tilt in radians

    Returns
    -------
    rotations_matrices : float[*, 3, 3]
        the rotation matrices
    """

    number_rotations = rotation.shape[0]
    rotations_matrices = np.zeros((number_rotations, 3, 3))

    for index in range(number_rotations):
        rotations_matrices[index, ...] = rotation_matrix_3D(tilt[index], 'y') @ rotation_matrix_3D(
            rotation[index], 'x'
        )
    return rotations_matrices


def ref_geometry(**kwargs):
    """
    create a reference geometry, based of default values and kwargs. The idea is not to cover any situation,
    but just enough to perform some test with this geometry on other functions

    Parameters
    ----------
    **kwrags : dict
        possible kwargs : length, thickness, volume, rotation_number, tilt_number, tilt_max.

    Returns
    -------
    geometry : Geometry
        the corresponding geometry
    """

    # load or default
    length = kwargs.get('length', 21)
    thickness = kwargs.get('thickness', 11)
    volume = kwargs.get(
        'volume', np.asarray([length, thickness, thickness])
    )  # rotation axis is here consider as being x aka 1
    rotation_number = kwargs.get(
        'rotation_number', thickness
    )  # by default enough to reconstruct the tomogram according to the usual rules of thumb
    tilt_number = kwargs.get('tilt_number', 3)
    tilt_max = kwargs.get('tilt_max', 15)  # in degree

    # complete information
    # corresponding to rotation_axis = 1
    projection_number = rotation_number * tilt_number
    projection_shape = np.asarray([volume[1], volume[0]])

    rotation = np.tile(np.linspace(0, 180, rotation_number), (tilt_number, 1)).flatten()
    tilt = np.tile(np.linspace(0, tilt_max, tilt_number)[..., np.newaxis], (1, rotation_number)).flatten()

    rotations = rot_matrices(rotation / 180 * 2 * np.pi, tilt / 180 * np.pi)

    j_offsets = np.zeros(projection_number)
    k_offsets = np.zeros(projection_number)
    p_direction_0 = np.asarray([0, 0, 1])
    j_direction_0 = np.asarray([0, 1, 0])
    k_direction_0 = np.asarray([1, 0, 0])
    detector_direction_origin = np.asarray([1, 0, 0])
    detector_direction_positive_90 = np.asarray([0, 1, 0])
    two_theta = 0.0
    detector_angles = np.asarray([0])

    # empty geometry object to complete
    geometry = Geometry()

    # adding general data
    geometry.p_direction_0 = p_direction_0
    geometry.j_direction_0 = j_direction_0
    geometry.k_direction_0 = k_direction_0
    geometry.detector_direction_origin = detector_direction_origin
    geometry.detector_direction_positive_90 = detector_direction_positive_90
    geometry.two_theta = two_theta
    geometry.detector_angles = detector_angles
    geometry.projection_shape = projection_shape
    geometry.volume_shape = volume

    # adding projections data
    for proj_index in range(projection_number):
        geom_tp = GeometryTuple(
            rotation=rotations[proj_index],
            j_offset=j_offsets[proj_index],
            k_offset=k_offsets[proj_index],
            inner_angle=rotation[proj_index],
            outer_angle=tilt[proj_index],
            inner_axis=np.array([1.0, 0.0, 0.0]),
            outer_axis=np.array([0.0, 1.0, 0.0]),
        )
        geometry.append(geom_tp)
    return geometry


def ref_data_container(tomogram, geometry):
    """
    create a DataContainer filled with the tomoragm projections associated with the given geometry

    Parameters
    ----------
    tomogram : float[*, *, *]
        A tomogram
    geometry : Geometry
        The accosicated Geometry object

    Returns
    -------
    data_container : DataContainer
        the DataContainer filled with the Projection object, based on the projections of the tomogram by a
        projector defined by the given geometry.

    """

    # empty data container to fill
    data_container = DataContainer()

    # projector related to the geometry
    projector = SAXSProjector(geometry)

    # compute the projections
    sinogram = projector.forward(tomogram)  # shape of sinogram == [proj number, x, y, 1]

    # normalization
    max_value = np.max(sinogram)
    sinogram = sinogram / max_value
    tomogram = tomogram / max_value

    # to have transmission and not absorption
    sinogram = 1 - sinogram

    # no masking, and standard weights
    weights = np.ones(sinogram.shape)

    # make projections with the tomogram and its associated geometry and save that a Projection object
    for proj_index in range(sinogram.shape[0]):
        projection = Projection(
            data=None,
            diode=sinogram[proj_index, :, :, 0],  # last 0 for the extra empty dim
            rotation=geometry.rotations[proj_index],
            j_offset=geometry.j_offsets[proj_index],
            k_offset=geometry.k_offsets[proj_index],
            inner_angle=geometry.inner_angles[proj_index],
            outer_angle=geometry.outer_angles[proj_index],
            inner_axis=geometry.inner_axes[proj_index],
            outer_axis=geometry.outer_axes[proj_index],
            weights=weights[proj_index],
        )

        # append this projection object to the DataContainer
        data_container.append(projection)
    # add general informations to the geometry in the DataContainer
    data_container.geometry.p_direction_0 = geometry.p_direction_0
    data_container.geometry.j_direction_0 = geometry.j_direction_0
    data_container.geometry.k_direction_0 = geometry.k_direction_0
    data_container.geometry.detector_direction_origin = geometry.detector_direction_origin
    data_container.geometry.detector_direction_positive_90 = geometry.detector_direction_positive_90
    data_container.geometry.two_theta = geometry.two_theta
    data_container.geometry.detector_angles = geometry.detector_angles
    data_container.geometry.projection_shape = geometry.projection_shape
    data_container.geometry.volume_shape = geometry.volume_shape

    return data_container


def test_mitra():
    """
    The idea is just to test if the mumott algo can reconstruct our data set
    """

    # defining reference parameters
    length = 21
    thickness = 11
    # in the case of rot_axis = 1

    # creating reference model
    geo = ref_geometry(length=length, thickness=thickness)
    tomo = ref_tomogram(geometry=geo)
    data_container = ref_data_container(tomo, geo)
    # normalization
    max_value = np.max(data_container.diode)
    tomo = tomo / max_value
    sino = 1 - np.moveaxis(data_container.diode, 0, -1)  # from transmission to absorption

    tomo_reconstructed = run_mitra(
        data_container, use_gpu=False, maxiter=50, ftol=None, enforce_non_negativity=True
    )['result']['x']
    # projector related to the geometry
    projector = SAXSProjector(geo)
    # compute the projections
    sino_reconstructed = projector.forward(tomo_reconstructed)  # shape of sinogram == [proj number, x, y, 1]
    sino_reconstructed = np.moveaxis(np.squeeze(sino_reconstructed), 0, -1)

    return sino, sino_reconstructed


def test_phase_matching_alignment_geometry_fom_data(**kwargs):
    """
    test the function that define the index of the geometrical y axis, i.e the main rotation axis from the
    data, and the related tomogram volume, from the DataContainer

    """

    # defining reference parameters
    length = 21
    thickness = 11
    rot_axis = 1
    volume_shape = np.asarray([length, thickness, thickness])

    # creating reference model
    geo = ref_geometry(length=length, thickness=thickness)
    tomo = ref_tomogram(geometry=geo)
    data_container = ref_data_container(tomo, geo)

    # extracting geometry information from it
    main_rot_axis_deduced, volume_deduced = get_alignment_geometry(data_container)

    # testing
    assert np.allclose(main_rot_axis_deduced, rot_axis)
    assert np.allclose(volume_deduced, volume_shape)


def test__define_axis_index():
    """
    y should be define as the rotation axis in the letter definitions. This function ensure that. This
    function is so simple that it might just need to be removed maybe.
    """
    # parameters definition
    main_rot_axis = 0
    # function to have x and y axis
    x_axis_index, y_axis_index = amo._define_axis_index(main_rot_axis)
    # test
    assert np.allclose(x_axis_index, 1)
    assert np.allclose(y_axis_index, 0)


def test_recenter_tomogram():
    """
    The tested function should recenter/move a tomogram toward its center of mass, at a given step rate
    (default 0.2) in the computed direction
    """

    # defining reference parameters
    length = 21
    thickness = 5
    # it is supposed that rot_axis = 1

    # creating reference model
    geo = ref_geometry(length=length, thickness=thickness)
    tomo = ref_tomogram(geometry=geo)

    # move by some pixels on the transervse plane in both direction
    tomo_moved = np.swapaxes(np.squeeze(tomo), 0, -1)
    tomo_moved = imp.imshift_fft(tomo_moved, 1 * np.ones(length), 0 * np.ones(length))
    tomo_moved = np.swapaxes(np.squeeze(tomo_moved), -1, 0)[..., np.newaxis]

    # center
    recentered_tomo = amo.recenter_tomogram(tomo_moved)

    # check dimensions
    assert np.allclose(recentered_tomo.shape, tomo_moved.shape)

    # check value
    # according to the tomogram definition, there is no data on first and last slice
    # the traverse slice were centered before the force shift of 1 and 0
    # the correction found is divided by 5 for a stepping approach
    # the the best recentering would be of 1 - 1/5 and 0
    tomo_recentered_ref = np.swapaxes(np.squeeze(tomo_moved), 0, -1)
    tomo_recentered_ref = imp.imshift_fft(tomo_recentered_ref,
                                          -1 * 0.2 * np.ones(length),
                                          -0 * 0.2 * np.ones(length))
    tomo_recentered_ref = np.swapaxes(np.squeeze(tomo_recentered_ref), -1, 0)[..., np.newaxis]

    # comparision
    assert np.allclose(recentered_tomo, tomo_recentered_ref, atol=0.3)


def test_compute_shift_through_sinogram():
    """
    Test the function that compute the shift needed to obtain a better sinogram, based on actual shifted
    sinogram and the sythetic sinogram.
    based on the sinogram from the data, the synthetic one, geometrical parameters (main rotation axis) and
    some parameters to ease the retrieval (filter, blocked axis ...)
    This function relies heavilly on the image processing code.

    The value to be found as shift is tricky to evaluate so its not tested here yet
    """

    # defining reference parameters
    length = 21
    thickness = 11
    rot_axis = 1

    # creating reference model
    geo = ref_geometry(length=length, thickness=thickness)
    tomo = ref_tomogram(geometry=geo)
    data_container = ref_data_container(tomo, geo)

    # normalization
    max_value = np.max(data_container.diode)
    tomo = tomo / max_value
    sino = 1 - np.moveaxis(data_container.diode, 0, -1)  # from transmission to absorption

    # create default parameters for the function to test
    x_axis_index, y_axis_index = amo._define_axis_index(rot_axis)
    param_high_pass_filter = 0.01  # High pass filter to emphasis details
    W = imp.compute_tukey_window(thickness, length)  # Tuckey Window of the size of each projection
    param_align_horizontal = True  # allow horizontal alignment (on X axis)
    param_align_vertical = True  # allow vertical alignment (on Y axis)

    if x_axis_index == 0:
        W = W.transpose(1, 0, 2)
    # shift the sinogram with a sinusoidal pattern over the whole data set
    shift_x = np.cos(np.linspace(0, 2 * np.pi, sino.shape[-1]))
    shift_y = np.sin(np.linspace(0, 2 * np.pi, sino.shape[-1]))
    sino_shifted = imp.imshift_fft(sino, shift_y, shift_x)

    # compute the related tomogram and its sythetized sinogram
    # update the shifts for SIRT
    for ii in range(data_container.projections.diode.shape[0]):
        data_container.projections[ii].diode = 1 - sino_shifted[..., ii]
    tomo_shifted = run_mitra(
        data_container, use_gpu=False, maxiter=50, ftol=None, enforce_non_negativity=True
    )['result']['x']
    projector = SAXSProjector(geo)
    synth_sino = projector.forward(tomo_shifted)  # shape of sinogram == [proj number, x, y, 1]
    synth_sino = np.squeeze(synth_sino)
    synth_sino = np.moveaxis(synth_sino, 0, -1)

    # compute the shifts
    shift_hor, shift_vect = amo.compute_shift_through_sinogram(
        sino_shifted,
        synth_sino,
        x_axis_index,
        y_axis_index,
        param_high_pass_filter,
        W,
        param_align_horizontal,
        param_align_vertical,
    )

    # compute blank shifts that should be 0
    shift_hor_0, shift_vect_0 = amo.compute_shift_through_sinogram(
        sino,
        sino,
        x_axis_index,
        y_axis_index,
        param_high_pass_filter,
        W,
        param_align_horizontal,
        param_align_vertical,
    )

    # check dim
    assert np.allclose(shift_hor.shape, shift_x.shape)
    assert np.allclose(shift_vect.shape, shift_y.shape)

    # check stability
    assert np.allclose(shift_hor_0, np.zeros(shift_x.shape))
    assert np.allclose(shift_vect_0, np.zeros(shift_y.shape))


def test_compute_shifts():
    """
    Test the function that is basically the core of the alignment procedure, where all the calculus and update
    are made. The full procedure is basically a loop over this functions.
    """

    # defining reference parameters
    length = 21
    thickness = 11
    rot_axis = 1

    # creating reference model
    geo = ref_geometry(length=length, thickness=thickness)
    tomo = ref_tomogram(geometry=geo)
    data_container = ref_data_container(tomo, geo)
    # normalization
    max_value = np.max(data_container.diode)
    tomo = tomo / max_value

    # get absorbance for the sino
    abs_dict = get_absorbances(data_container.diode, normalize_per_projection=True)
    absorbances = abs_dict['absorbances']
    sinogram_0 = np.moveaxis(np.squeeze(absorbances), 0, -1)

    # reference for the diodes
    diodes_ref = np.moveaxis(np.squeeze(data_container.diode), 0, -1)

    # create default parameters for the function to test
    x_axis_index, y_axis_index = amo._define_axis_index(rot_axis)
    param_high_pass_filter = 0.01  # High pass filter to emphasis details
    W = imp.compute_tukey_window(thickness, length)  # Tuckey Window of the size of each projection
    param_align_horizontal = True  # allow horizontal alignment (on X axis)
    param_align_vertical = True  # allow vertical alignment (on Y axis)
    param_center_reconstruction = True
    param_rec_iteration = 50

    if x_axis_index == 0:
        W = W.transpose(1, 0, 2)
    # shift the sinogram with a sinusoidal pattern over the whole data set
    shift_v = np.cos(np.linspace(0, 2 * np.pi, sinogram_0.shape[-1]))
    shift_h = np.sin(np.linspace(0, 2 * np.pi, sinogram_0.shape[-1]))
    shift_all = np.concatenate((shift_v[..., np.newaxis], shift_h[..., np.newaxis]), axis=1)[
        np.newaxis, ...
    ]  # first dim to store the iterations
    shift_all = np.concatenate((shift_all, shift_all), axis=0)

    # shift the original diodes
    diodes_shifted = imp.imshift_fft(np.moveaxis(np.squeeze(data_container.diode), 0, -1), shift_v, shift_h)
    # compute the related tomogram and its sythetized sinogram
    # update the shifts for SIRT
    for ii in range(data_container.projections.diode.shape[0]):
        data_container.projections[ii].diode = diodes_shifted[..., ii]
    # lets test the function now
    correction_criterion, sinogram_corr, tomogram, shift_all_res = amo.compute_shifts(
        sinogram_0,
        diodes_ref,
        shift_all,
        data_container,
        0,
        x_axis_index,
        y_axis_index,
        param_rec_iteration,
        param_high_pass_filter,
        W,
        param_align_horizontal,
        param_align_vertical,
        param_center_reconstruction,
        False,
        dict(),
    )

    sinogram_corr = np.moveaxis(sinogram_corr, 0, -1)

    # check dimensions
    assert np.allclose(sinogram_corr.shape, sinogram_0.shape)
    assert np.allclose(tomogram.shape, tomo.shape)
    assert np.allclose(shift_all_res.shape, shift_all.shape)

    # verify that there is no NaN seems the best way to check
    assert np.allclose(np.isnan(shift_all_res), np.isnan(shift_all))


def test_optical_flow_alignment(return_result: bool = False):
    """
    Test the full procedure.
    """

    # defining reference parameters
    length = 21
    thickness = 11
    # in the case of rot_axis = 1

    # creating reference model
    geo = ref_geometry(length=length, thickness=thickness)
    tomo = ref_tomogram(geometry=geo)
    data_container = ref_data_container(tomo, geo)
    # normalization
    max_value = np.max(data_container.diode)
    tomo = tomo / max_value

    # get absorbance for the sino
    abs_dict = get_absorbances(data_container.diode, normalize_per_projection=True)
    absorbances = abs_dict['absorbances']
    sino = np.moveaxis(np.squeeze(absorbances), 0, -1)

    # shift the sinogram with a sinusoidal pattern over the whole data set
    shift_x = (
        np.cos(np.linspace(0, 2 * np.pi, sino.shape[-1])) * 0
    )  # we dont want ni line shift here, due to our sample shape
    shift_y = np.sin(np.linspace(0, 2 * np.pi, sino.shape[-1])) * 1  # transversal shift

    # compute the related tomogram and its sythetized sinogram
    # update the shifts for SIRT
    # shift the original diodes
    diodes_shifted = imp.imshift_fft(np.moveaxis(np.squeeze(data_container.diode), 0, -1), shift_y, shift_x)

    # update the shifts for SIRT / MITRA
    for ii in range(data_container.projections.diode.shape[0]):
        data_container.projections[ii].diode = diodes_shifted[..., ii]

    # get shifted absorbance for the shiftedv sino
    abs_dict_shifted = get_absorbances(data_container.diode, normalize_per_projection=True)
    absorbances_shifted = abs_dict_shifted['absorbances']
    sino_shifted = np.moveaxis(np.squeeze(absorbances_shifted), 0, -1)

    # run with default
    shift_all, sinogram_corr, tomogram = amo.run_optical_flow_alignment(
        data_container, stop_max_iteration=10, rec_iteration=10, center_reconstruction=True
    )

    # realign sino that was shifted originaly
    sino_found_shifted = imp.imshift_fft(sino_shifted, shift_all[..., 0], shift_all[..., 1])

    # check dimensions
    assert np.allclose(sinogram_corr.shape, sino.shape)
    assert np.allclose(tomogram.shape, tomo.shape)

    # verify that there is no NaN seems the best way to check
    assert np.allclose(np.isnan(shift_all), np.isnan(np.zeros(shift_all.shape)))

    # verify value of almost perfect alignment
    accepted_error = 2.5
    assert np.allclose(sino, sino_found_shifted, atol=accepted_error)

    if return_result:
        return sino, sino_shifted, sinogram_corr, sino_found_shifted


def test_sinogram_line_vertical_alignment(return_result: bool = False):
    """
    Test line vertical alignment function that align a sinogram based of the center of mass
    """

    # defining reference parameters
    length = 21
    thickness = 11
    # in the case of rot_axis = 1

    # creating reference model
    geo = ref_geometry(length=length, thickness=thickness)
    tomo = ref_tomogram(geometry=geo)
    data_container = ref_data_container(tomo, geo)
    # normalization
    max_value = np.max(data_container.diode)
    tomo = tomo / max_value

    # get absorbance for the sino
    abs_dict = get_absorbances(data_container.diode, normalize_per_projection=True)
    absorbances = abs_dict['absorbances']
    sino = np.moveaxis(np.squeeze(absorbances), 0, -1)

    # shift the sinogram with a sinusoidal pattern over the whole data set
    shift_x = (
        np.cos(np.linspace(0, 2 * np.pi, sino.shape[-1])) * 0
    )  # we dont want ni line shift here, due to our sample shape
    shift_y = np.sin(np.linspace(0, 2 * np.pi, sino.shape[-1])) * 1  # transversal shift

    # compute the related tomogram and its sythetized sinogram
    # update the shifts for SIRT
    # shift the original diodes
    diodes_shifted = imp.imshift_fft(np.moveaxis(np.squeeze(data_container.diode), 0, -1), shift_y, shift_x)

    # update the shifts for SIRT / MITRA
    for ii in range(data_container.projections.diode.shape[0]):
        data_container.projections[ii].diode = diodes_shifted[..., ii]

    # get shifted absorbance for the shiftedv sino
    abs_dict_shifted = get_absorbances(data_container.diode, normalize_per_projection=True)
    absorbances_shifted = abs_dict_shifted['absorbances']
    sino_shifted = np.moveaxis(np.squeeze(absorbances_shifted), 0, -1)

    # run with default
    main_rotation_axis = 1
    vertical_shift, vertical_axis = amo.sinogram_line_vertical_alignment(sino_shifted, main_rotation_axis)

    # realign sino that was shifted originaly
    shift_all = np.zeros((sino.shape[-1], 2))
    shift_all[:, vertical_axis] = shift_all[:, vertical_axis] + vertical_shift
    sino_found_shifted = imp.imshift_fft(sino_shifted, shift_all[..., 0], shift_all[..., 1])

    # verify that there is no NaN seems the best way to check
    assert np.allclose(np.isnan(shift_all), np.isnan(np.zeros(shift_all.shape)))

    # verify value of almost perfect alignment
    accepted_error = 2.5
    assert np.allclose(sino, sino_found_shifted, atol=accepted_error)

    if return_result:
        return sino, sino_shifted, sino_found_shifted, vertical_shift


def test_line_vertical_alignment(return_result: bool = False):
    """
    Test line vertical alignment function for DataContainer that align a sinogram based of the center of mass
    """

    # defining reference parameters
    length = 21
    thickness = 11
    # in the case of rot_axis = 1

    # creating reference model
    geo = ref_geometry(length=length, thickness=thickness)
    tomo = ref_tomogram(geometry=geo)
    data_container = ref_data_container(tomo, geo)
    # normalization
    max_value = np.max(data_container.diode)
    tomo = tomo / max_value

    # get absorbance for the sino
    abs_dict = get_absorbances(data_container.diode, normalize_per_projection=True)
    absorbances = abs_dict['absorbances']
    sino = np.moveaxis(np.squeeze(absorbances), 0, -1)

    # shift the sinogram with a sinusoidal pattern over the whole data set
    shift_x = (
        np.cos(np.linspace(0, 2 * np.pi, sino.shape[-1])) * 0
    )  # we dont want ni line shift here, due to our sample shape
    shift_y = np.sin(np.linspace(0, 2 * np.pi, sino.shape[-1])) * 1  # transversal shift

    # compute the related tomogram and its sythetized sinogram
    # update the shifts for SIRT
    # shift the original diodes
    diodes_shifted = imp.imshift_fft(np.moveaxis(np.squeeze(data_container.diode), 0, -1), shift_y, shift_x)

    # update the shifts for SIRT / MITRA
    for ii in range(data_container.projections.diode.shape[0]):
        data_container.projections[ii].diode = diodes_shifted[..., ii]

    # get shifted absorbance for the shiftedv sino
    abs_dict_shifted = get_absorbances(data_container.diode, normalize_per_projection=True)
    absorbances_shifted = abs_dict_shifted['absorbances']
    sino_shifted = np.moveaxis(np.squeeze(absorbances_shifted), 0, -1)

    vertical_shift = amo.line_vertical_alignment(data_container)
    shift_all = vertical_shift

    # realign sino that was shifted originaly
    sino_found_shifted = imp.imshift_fft(sino_shifted, shift_all[..., 0], shift_all[..., 1])

    # verify that there is no NaN seems the best way to check
    assert np.allclose(np.isnan(shift_all), np.isnan(np.zeros(shift_all.shape)))

    # verify value of almost perfect alignment
    accepted_error = 2.5
    assert np.allclose(sino, sino_found_shifted, atol=accepted_error)

    if return_result:
        return sino, sino_shifted, sino_found_shifted, vertical_shift


def test_shifts_to_vector():
    """
    Test the shifts to sensor geometry function.
    """
    # defining reference parameters
    length = 21
    thickness = 11

    geo = ref_geometry(length=length, thickness=thickness)
    tomo = ref_tomogram(geometry=geo)
    data_container = ref_data_container(tomo, geo)
    main_rot_axis_deduced, _ = get_alignment_geometry(data_container)
    other_axis = 0 if main_rot_axis_deduced == 1 else 1

    # shift the sinogram with a sinusoidal pattern over the whole data set
    shift_x = (
        np.cos(np.linspace(0, 2 * np.pi, 33)) * 0.1
    )  # we dont want ni line shift here, due to our sample shape
    shift_y = np.sin(np.linspace(0, 2 * np.pi, 33)) * 1  # transversal shift
    shifts = np.concatenate((shift_x[:, np.newaxis], shift_y[:, np.newaxis]), axis=1)

    vector_shifts = amo.shifts_to_vector(shifts, data_container)

    # verify that the new coordinates are good
    inner_index = np.where(np.isclose(np.abs(geo[0].inner_axis), 1))[0][0]
    outer_index = np.where(np.isclose(np.abs(geo[0].outer_axis), 1))[0][0]

    # y correspond to the inner index
    assert np.allclose(shifts[:, main_rot_axis_deduced], vector_shifts[:, inner_index])
    assert np.allclose(shifts[:, other_axis], vector_shifts[:, outer_index])


def test_shifts_to_geometry(return_result: bool = False):
    """
    Test the shifts to object geometry function.
    """

    # defining reference parameters
    length = 21
    thickness = 11
    # in the case of rot_axis = 1

    # creating reference model
    geo = ref_geometry(length=length, thickness=thickness)
    tomo = ref_tomogram(geometry=geo)
    data_container = ref_data_container(tomo, geo)
    # normalization
    max_value = np.max(data_container.diode)
    tomo = tomo / max_value

    # get absorbance for the sino
    abs_dict = get_absorbances(data_container.diode, normalize_per_projection=True)
    absorbances = abs_dict['absorbances']
    sino = np.moveaxis(np.squeeze(absorbances), 0, -1)

    ref_tomo = run_mitra(
        data_container, use_gpu=False, maxiter=50, ftol=None, enforce_non_negativity=True
    )['result']['x']

    # shift the sinogram with a sinusoidal pattern over the whole data set
    shift_x = (
        np.cos(np.linspace(0, 2 * np.pi, sino.shape[-1])) * 1
    )  # we dont want ni line shift here, due to our sample shape
    shift_y = np.sin(np.linspace(0, 2 * np.pi, sino.shape[-1])) * 1  # transversal shift

    # apply shifts
    data_container.geometry.j_offsets = shift_x
    data_container.geometry.k_offsets = shift_y

    shifted_tomo = run_mitra(
        data_container, use_gpu=False, maxiter=50, ftol=None, enforce_non_negativity=True
    )['result']['x']

    # run with default
    shift_all, _, _ = amo.run_optical_flow_alignment(
        data_container, stop_max_iteration=10, rec_iteration=10, center_reconstruction=True
    )

    # now lets check the mapping of the geometry between the projection shifts and the j-k space
    j_offsets, k_offsets = amo.shifts_to_geometry(data_container, shift_all)
    data_container.geometry.j_offsets += j_offsets
    data_container.geometry.k_offsets += k_offsets

    tomo_reconstructed_innershifts = run_mitra(
        data_container, use_gpu=False, maxiter=50, ftol=None, enforce_non_negativity=True
    )['result']['x']

    # verify value of almost perfect alignment
    accepted_error = 5
    assert np.allclose(shift_x, -j_offsets, atol=accepted_error)
    assert np.allclose(shift_y, -k_offsets, atol=accepted_error)

    if return_result:
        return ref_tomo, shifted_tomo, tomo_reconstructed_innershifts
