import numpy as np
from mumott.core import calculate_sph_coefficients_rotated_by_euler_angles, load_d_matrices
from mumott.core.wigner_d_utilities import (
    calculate_sph_coefficients_rotated_around_z,
    calculate_sph_coefficients_rotated_around_z_derived_wrt_the_angle,
    )
from mumott.methods.utilities.fiber_fit import (
    find_approximate_symmetry_axis,
    symmetric_part_along_given_direction,
    degree_of_symmetry_map,
    )

rotated_coeffs_expected = np.array([1., 0.55758988, 0.66263709, -0.06211013, 0.42547438,
                                    -0.25518522, -0.2805971, 0.09502369, 0.37557435, -0.28929866,
                                    -0.34688224, -0.18575653, -0.17188444, -0.66661516, -0.24234923,
                                    -0.06662794, -0.5086025, -0.41621331, 0.00868982, -0.42055614,
                                    -0.20720317, 0.28465786, -0.13304362, 0.19247069, -0.06096128,
                                    -0.35947975, 0.15045119, 0.22895712])

coeffs_test = np.array([1,
                        0, 0, 1, 0, 0,
                        0, 0, 0, 0, 1, 0, 0, 0, 0,
                        0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0]).reshape((1, 1, 1, 28))

exp_map = np.array([[1.2144377, 1.2144377, 1.2144377, 1.2144377, 1.2144377,
                     1.2144377, 1.2144377, 1.2144377, 1.2144377, 1.2144377],
                    [0.8973241, 2.28580287, 1.31798271, 0.75870596, 0.26787511,
                     1.31565715, 0.39927551, 1.54529699, 0.32599202, 0.8973241],
                    [1.19995005, 1.83236137, 1.60223676, 0.7911506, 0.56414813,
                     0.76468222, 1.87134966, 0.85012911, 1.19492182, 1.19995005],
                    [1.51677032, 0.55724607, 0.6988529, 1.17611535, 1.28119196,
                     1.76659705, 6.09947232, 1.89801435, 1.21137914, 1.51677032],
                    [1.2144377, 1.2144377, 1.2144377, 1.2144377, 1.2144377,
                     1.2144377, 1.2144377, 1.2144377, 1.2144377, 1.2144377]])


def test_loading():
    d = load_d_matrices(9)
    assert np.allclose(d[8]@d[8].T, np.identity(17))


def test_rotation():
    coeffs_test_local = np.ones((2, 3, 3, 28)) * coeffs_test
    theta = 1
    phi = 1
    rotated_coeffs = calculate_sph_coefficients_rotated_by_euler_angles(coeffs_test_local, None, theta, phi)
    # print(np.array_repr(rotated_coeffs[0, 0, 0, :]))
    assert np.allclose(rotated_coeffs[0, 0, 0, :], rotated_coeffs_expected)


def test_differentiation():

    coeffs_test_local = np.ones((2, 3, 3, 28)) * coeffs_test
    theta = 2.1
    phi = 3
    rotated_coeffs = calculate_sph_coefficients_rotated_by_euler_angles(coeffs_test_local, None, theta, phi)
    psi = 1
    del_psi = 0.01
    fd_deriv = calculate_sph_coefficients_rotated_around_z(rotated_coeffs,
                                                           np.ones((2, 3, 3)) * (psi+del_psi),
                                                           np.arange(0, 6+1, 2))
    fd_deriv -= calculate_sph_coefficients_rotated_around_z(rotated_coeffs,
                                                            np.ones((2, 3, 3)) * psi,
                                                            np.arange(0, 6+1, 2))
    fd_deriv = fd_deriv / del_psi
    analy_deriv = calculate_sph_coefficients_rotated_around_z_derived_wrt_the_angle(rotated_coeffs,
                                                                                    psi * np.ones((2, 3, 3)),
                                                                                    np.arange(0, 6+1, 2))
    assert np.allclose(fd_deriv, analy_deriv, atol=0.05, rtol=0.02)


def test_fiberfit():

    rot_coeffs = rotated_coeffs_expected.reshape((1, 1, 1, rotated_coeffs_expected.shape[0]))
    coeffs, theta, phi = find_approximate_symmetry_axis(rot_coeffs,
                                                        ell_max=6,
                                                        resolution=40)

    assert np.allclose(coeffs[0, 0, 0, :], np.ones(4),  rtol=0.02)
    assert np.allclose(theta[0, 0, 0], np.ones(1),  rtol=0.05)
    assert np.allclose(phi[0, 0, 0], np.ones(1),  rtol=0.05)

    dos_map, _, _ = degree_of_symmetry_map(rot_coeffs,
                                           ell_max=6,
                                           resolution=5,
                                           filter='square')
    assert np.allclose(dos_map, exp_map)

    coeffs = symmetric_part_along_given_direction(rot_coeffs,
                                                  np.ones((1, 1, 1)),
                                                  np.ones((1, 1, 1)),
                                                  ell_max=6)
    assert np.allclose(coeffs[0, 0, 0, :], np.ones(4))
