import pytest # noqa
import numpy as np
from mumott.data_handling import DataContainer
from mumott.pipelines.async_pipelines import (run_tensor_sirt,
                                              run_motr,
                                              run_radtt)
from mumott.methods.utilities.grids_on_the_sphere import football
from mumott.methods.basis_sets import NearestNeighbor


@pytest.fixture
def data_container():
    return DataContainer('tests/test_half_circle.h5')


expected_stsrt = np.array([[5.385921, 5.369867, 0., 0., 0., 0.,
                            0., 0., 5.520075, 5.527561, 4.3478236, 4.3070016,
                            0., 0., 5.051396, 5.0786176, 4.1699095, 4.199969],
                           [3.7770133, 3.7897654, 0., 0., 0., 0.,
                            0., 0., 4.618585, 4.5898366, 4.524088, 4.5688133,
                            0., 0., 6.076858, 6.055185, 5.6179967, 5.657434],
                           [3.690018, 3.7171938, 0., 0., 0., 0.,
                            0., 0., 4.0379896, 4.0114884, 5.400281, 5.4768457,
                            0., 0., 5.4900417, 5.4439306, 6.2792196, 6.2667365],
                           [3.447556, 3.4897063, 0., 0., 0., 0.,
                            0., 0., 3.5942144, 3.5625627, 6.132254, 6.2459064,
                            0., 0., 5.3961186, 5.324624, 7.1005783, 7.054971]], dtype=np.float32)


def test_tensor_sirt(data_container):
    result = run_tensor_sirt(data_container)
    print(repr(result['reconstruction'][2, 2, ...]))
    assert np.allclose(result['reconstruction'][2, 2], expected_stsrt)


expected_motr = np.array([[0.0000000e+00, 0.0000000e+00, 0.0000000e+00, 0.0000000e+00,
                           0.0000000e+00, 0.0000000e+00, 0.0000000e+00, 0.0000000e+00,
                           0.0000000e+00, 0.0000000e+00, 0.0000000e+00, 0.0000000e+00,
                           0.0000000e+00, 0.0000000e+00, 0.0000000e+00, 0.0000000e+00,
                           0.0000000e+00, 0.0000000e+00],
                          [9.669315, 9.702539, 0., 0., 0., 0.,
                           0., 0., 11.680179, 11.609665, 11.630562, 11.74468,
                           0., 0., 15.269386, 15.21302, 14.29571, 14.386248],
                          [9.425116, 9.496931, 0., 0., 0., 0.,
                           0., 0., 10.200384, 10.133834, 13.958943, 14.159476,
                           0., 0., 13.869945, 13.748242, 16.133106, 16.090212],
                          [0.0000000e+00, 0.0000000e+00, 0.0000000e+00, 0.0000000e+00,
                           0.0000000e+00, 0.0000000e+00, 0.0000000e+00, 0.0000000e+00,
                           0.0000000e+00, 0.0000000e+00, 0.0000000e+00, 0.0000000e+00,
                           0.0000000e+00, 0.0000000e+00, 0.0000000e+00, 0.0000000e+00,
                           0.0000000e+00, 0.0000000e+00]], dtype=np.float32)


def test_motr(data_container):
    result = run_motr(data_container, lower_bound=0.)
    print(repr(result['reconstruction'][2, 2, ...]))
    assert np.allclose(result['reconstruction'][2, 2], expected_motr)


expected_football = np.array([[0., 0., 0., 0., 0., 0.,
                               0., 0., 0., 0., 0., 0.,
                               0., 0., 0., 0.],
                              [0., 0., 10.672098, 0., 0., 0.,
                               11.245457, 12.074425, 0., 0., 0., 0.,
                               0., 0., 0., 0.],
                              [0., 0., 10.220264, 0., 0., 0.,
                               12.649972, 11.618044, 0., 0., 0., 0.,
                               0., 0., 0., 0.],
                              [0., 0., 0., 0., 0., 0.,
                               0., 0., 0., 0., 0., 0.,
                               0., 0., 0., 0.]], dtype=np.float32)


def test_motr_football(data_container):
    bs = NearestNeighbor(football(),
                         probed_coordinates=data_container.geometry.probed_coordinates,
                         integration_mode='midpoint')
    result = run_motr(data_container, basis_set=bs, lower_bound=0.)
    print(repr(result['reconstruction'][2, 2, ...]))
    assert np.allclose(result['reconstruction'][2, 2], expected_football)


expected_football_tensor_sirt = np.array([[0., 0., 4.82039, 0., 0., 0.,
                                           3.7433403, 4.238179, 0., 0., 0., 0.,
                                           0., 0., 0., 0.],
                                          [0., 0., 4.0103827, 0., 0., 0.,
                                           4.2165484, 4.7206864, 0., 0., 0., 0.,
                                           0., 0., 0., 0.],
                                          [0., 0., 3.8870215, 0., 0., 0.,
                                           4.696129, 4.4811983, 0., 0., 0., 0.,
                                           0., 0., 0., 0.],
                                          [0., 0., 3.753602, 0., 0., 0.,
                                           5.200454, 4.507147, 0., 0., 0., 0.,
                                           0., 0., 0., 0.]], dtype=np.float32)


def test_tensor_sirt_football(data_container):
    bs = NearestNeighbor(football(),
                         probed_coordinates=data_container.geometry.probed_coordinates,
                         integration_mode='midpoint')
    result = run_tensor_sirt(data_container, basis_set=bs)
    print(repr(result['reconstruction'][2, 2, ...]))
    assert np.allclose(result['reconstruction'][2, 2], expected_football_tensor_sirt)


expected_radtt = np.array([[0., 0., 0., 0., 0.,
                            0., 0., 0., 0., 0.,
                            0., 0., 0., 0., 0.,
                            0., 0., 0.],
                           [10.643478, 10.654295, 0., 0., 0.,
                            0., 0., 0., 11.087453, 11.069557,
                            11.299245, 11.333681, 0., 0., 12.015302,
                            11.996931, 11.9559965, 11.971425],
                           [10.601709, 10.618052, 0., 0., 0.,
                            0., 0., 0., 10.872646, 10.855226,
                            11.6252775, 11.672127, 0., 0., 11.816583,
                            11.788846, 12.215851, 12.212458],
                           [0., 0., 0., 0., 0.,
                            0., 0., 0., 0., 0.,
                            0., 0., 0., 0., 0.,
                            0., 0., 0.]], dtype=np.float32)


def test_radtt(data_container):
    result = run_radtt(data_container)
    print(repr(result['reconstruction'][2, 2, ...]))
    assert np.allclose(result['reconstruction'][2, 2], expected_radtt)
